// Copyright (c) 2019, Cthulhu
//
// Redistribution and use in source and binary forms,
// with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
//    this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

arch snes.cpu

constant null(0)

// Boolean constants
constant false(0)
constant true(1)

constant inf($FF) // Infinite

macro seek(variable address) {
    variable bank((address >> 16) & $FF)
    variable offset(address & $FFFF)

    if {defined EX} {
        if (bank >= $00 && bank <= $3F) || (bank >= $40 && bank <= $7D) {
            // ExHiROM section
            // No FastROM support.
            if bank < $40 && offset < $8000 {
                error "Invalid ExHiROM address for seek."
            }

            origin address | $400000
            base address
        } else if (bank >= $80 && bank <= $BF) || (bank >= $C0 && bank <= $FF) {
            // HiROM section
            // FastROM support.
            if bank < $C0 && offset < $8000 {
                error "Invalid HiROM address for seek."
            }

            origin address & $3FFFFF
            base address
        } else {
            error "Invalid ExHiROM address for seek."
        }
    } else {
        if (bank >= $00 && bank <= $3F) || (bank >= $80 && bank <= $BF) {
            if offset < $8000 {
                error "Invalid HiROM address for seek."
            }

            origin address & $3FFFFF
            base address | $800000 // Move address to FastROM address.
        } else if (bank >= $40 && bank <= $7D) || (bank >= $C0 && bank <= $FF) {
            origin address & $3FFFFF
            base address | $800000 // Move address to FastROM address.
        } else {
            error "Invalid HiROM address for seek."
        }
    }
}

macro assert(string name, variable max) {
    if pc() > max {
        variable size(pc() - max)
        print "data: {name} overran by ",size," bytes.\n"
        error "Address space overrun!"
    }
}

macro return(variable source, variable offset) {
    // Return after jml.
    jml source+4+offset // jml length is 4.
}

include "charm.inc"
include "drops.inc"
include "catfood.inc"

// ========================
// = Original code hijack =
// ========================
scope Hooks {
    seek($C1FD8D)
    DROPS_CHECK:

    seek($C1FD7C)
    CHARM_CHECK:
        bra DROPS_CHECK // Skip charm slot check when initializing the loot table.

        while pc() < DROPS_CHECK {
            nop
        }

    seek($C1D176)
    HIT_CHECK:
        jml Patch.HitCheck

    seek($FDAC6D)
    END:
    // rtl
}

scope Patch {
seek($FDAC08)

scope EnemyDrops: {
    // This hack allows for more than 3 items to be dropped per battle and
    // repurposes the charm slot into the maximum number of the same item
    // that can be dropped per battle.

    lda.l $CC5E04,x // Enemy rewards (drop).
    beq Hooks.END
    sta.b $10

    lda.l $CC5E05,x // Enemy rewards (charm).

    sep #$10

    ldy.b #$00
-
    ldx.w $B2A7,y // Memory location containing the dropped items.
    beq .Drop
    iny

    cpx.b $10
    bne -
    cmp.b #1 // No more items of this type should drop?
    dec
    bcs -

.NoDrop:
    rep #$10

    rtl

.Drop:
    lda.b $10
    sta $B2A7,y

    lda.b #$20
    tsb.w $B2AF

    rep #$10

    rtl
}

scope HitCheck: {
    // This hack makes 1/2 HP damage from Twin Charm
    // only hit targets vulnerable to it.

    stz.b $17
    stz.b $19

    lda.w $B2C7
    bne .Default

    ldx.w $B2C3  // Tech
    cpx.w #$037B // Twin Charm
    bne .Default

    lda.b #$19 // Tech Immunities
    rep #$20
    clc
    adc.w $B1F6 // Target
    tax
    sep #$20

    lda.b #$FF

    bit.w $5E2D,x
    bpl .Hit // 1/2 HP damage and instant death

.Miss:
    stz.b $16
    sta.b $18

    bra .Exit

.Hit:
    sta.b $16
    stz.b $18

.Exit:
    return(Hooks.HIT_CHECK, $25)

.Default:
    return(Hooks.HIT_CHECK, 0)
}

assert("Patch", Hooks.END)

fill (Hooks.END - pc()),$EA // Unused bytes to nop.
}
